/*
 *     RenameField.java  Aug 15, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.fields;
import org.jtool.jrbx.refactor.methods.RenameMethod;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.platform.xml.JXConstants;
import org.jtool.platform.xml.wrapper.*;
import java.util.*;
import org.w3c.dom.*;
import javax.swing.JFrame;

public class RenameField extends FieldRefactoring {
    private String newName = null;
    private RenameFieldTransformer transformer;
    private boolean updateReferences = true;
    private JavaMethod getter;
    private JavaMethod setter;
    private RenameMethod grenamer;
    private RenameMethod srenamer;

    public RenameField() {
        super();
    }

    public void setUpWithOptions(JFrame f, String newName, boolean ref) {
        frame = f;
        this.newName = newName;
        updateReferences = ref;
    }

    public void preconditions() throws RefactoringException {
        getter = QueryField.guessGettter(jfield);
        setter = QueryField.guessSettter(jfield);

        if (newName == null) {
            String getterStr = "";
            String setterStr = "";
            if (getter != null) {
                getterStr = getter.getSignature();
            }
            if (setter != null) {
                setterStr = setter.getSignature();
            }
            RenameFieldDialog dialog = new RenameFieldDialog(frame, "Rename Field", jfield.getName(),
                                                              getterStr, setterStr);
            newName = dialog.getNewName();
            if (!dialog.getGetterCheckBox()) {
                getter = null;
            }
            if (!dialog.getSetterCheckBox()) {
                setter = null;
            }
            updateReferences = dialog.getUpdateCheckBox();

            if (newName == null) {
                throw new RefactoringException("");
            }
        }

        checkAlreadyUsed();
        checkInnerClass();
        checkOuterClass();
        checkLocalVariable();
        checkAccessors();
    }

    public void transform() throws RefactoringException {
        transformer = new RenameFieldTransformer(jfield, newName);
        transformer.changeSrc();
        unparse(jfile);
    }
    
    public void additionalTransformation() throws RefactoringException {
        if (!updateReferences) {
            return;
        }

        ArrayList files = QueryProject.getAllJavaFilesUsingTheField(jfield);
        Iterator it = files.iterator();
        while (it.hasNext()) {
            JavaFile jf = (JavaFile)it.next();
            transformer.changeRef(jf);
            unparse(jf);
        }

        changeAccessor();
        unparse(jfile);
    }

    private void changeAccessor() throws RefactoringException {
        if (getter != null) {
            grenamer.transform();
            grenamer.additionalTransformation();
        }
        if (setter != null) {
            srenamer.transform();
            srenamer.additionalTransformation();
        }
    }

    public String getLog() {
        return jfield.getName() + " to " + newName;
    }

    private void checkAlreadyUsed() throws RefactoringException {
        if (QueryClass.hasTheField(jclass, newName)) {
            throw new RefactoringException("Not rename:" +
              " the field " + newName + " already exists in the class " + jclass.getName());
        }
    }
    
    private void checkInnerClass() throws RefactoringException {
        Iterator it = jclass.getAllInnerClassesAndInterfaces().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            JavaField jf = jc.getJavaField(newName);
            if (jf != null) {
                throw new RefactoringException("Not rename:" +
                  " the field " + newName + " already exists in the class " + jc.getName());
            }
        }
    }

    private void checkOuterClass() throws RefactoringException {
        Iterator it = jclass.getAllOuterClasses().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            JavaField jf = jc.getJavaField(newName);
            if (jf != null) {
                throw new RefactoringException("Not rename:" +
                  " the field " + newName + " already exists in the class " + jc.getName());
            }
        }
    }

    private void checkLocalVariable() throws RefactoringException {
        Iterator it = QueryFile.getJavaVariablesUsingTheField(jfile, jfield).iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            if (jv.getDotPrefix() == null &&
                hasTheParameter(jv.getJavaMethod(), newName) || hasTheLocalVariableInScope(jv, newName)) {
                throw new RefactoringException("Not rename:" +
                  " the local variable/parameter " + newName + " already exists in the method " + jv.getJavaMethod().getName());
            }
        }
    }

    private boolean hasTheParameter(JavaMethod jm, String varName) {
        return jm.getParameterNode(varName) != null;
    }

    private boolean hasTheLocalVariableInScope(JavaVariable jv, String varName) {
        Node node = jv.getElem();
        while (node != null) {
            if (hasTheLocalVariableInScope(node, varName) && jv.getDotPrefix() == null) {
                return true;
            }

            if (node.getNodeName().equals(JXConstants.MethodElem) ||
                node.getNodeName().equals(JXConstants.StaticInitElem)) {
                return false;
            }
            node = node.getParentNode();
        }
        return false;
    }

    private boolean hasTheLocalVariableInScope(Node node, String varName) {
        while (node != null) {
            if (node.getNodeName().equals(JXConstants.LocalElem)) {
                JavaLocal jl = new JavaLocal((Element)node);
                if (jl.getName().equals(varName)) {
                    return true;
                }
            }
            node = node.getPreviousSibling();
        }
        return false;
    }

    private void checkAccessors() {
        if (getter != null) {
            try {
                grenamer = new RenameMethod();
                grenamer.setUp(getter.getElem());
                grenamer.setUpWithOptions(frame, "get" + QueryField.getSuffixName(newName), updateReferences);
                grenamer.preconditions();
            } catch (RefactoringException e) {
                getter = null;
            }
        }
        
        if (setter != null) {
            try {
                srenamer = new RenameMethod();
                srenamer.setUp(setter.getElem());
                srenamer.setUpWithOptions(frame, "set" + QueryField.getSuffixName(newName), updateReferences);
                srenamer.preconditions();
            } catch (RefactoringException e) {
                setter = null;
            }
        }
    }
}
