/*
 *     MoveMethod.java  Aug 28, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.methods;
import org.jtool.platform.xml.JXConstants;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.j2xml.SystemProperties;
import java.util.*;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import javax.swing.JFrame;

public class MoveMethod extends MethodRefactoring {
    private JavaClass dclass;
    private String refName;
    private boolean deleteParam = false;
    private boolean insertParam = false;
    private MoveMethodTransformer transformer;

    public MoveMethod() {
        super();
    }

    public void setUpWithOptions(JFrame f, JavaClass dclass, String refName) {
        frame = f;
        this.dclass = dclass;
        this.refName = refName;
    }

    public void preconditions() throws RefactoringException {
        checkAbstract();
        checkNative();
        checkSynchronized();
        checkCallWithSuper();
        checkThisSuffix();
        checkOverriddenOrOverridingMethod();
        if (!jmethod.isStatic()) {
            checkParameterAndField();
        } else {
            checkMethodCall();
            checkFieldAccess();
        }

        if (dclass == null) {
            MoveMemberDialog dialog = new MoveMemberDialog(frame, "Move Method: " + jmethod.getName(), jclass.getFqn(),
                             SystemProperties.getProjectDir(), QueryProject.getAllJavaClassesInProject().getArrayList());
            dclass = dialog.getJavaClass();

            if (dclass == null) {
                throw new RefactoringException("");
            }
        }

        checkTargetClass();
        checkMovedMethod();
        determineReferenceName();
    }

    public void transform() throws RefactoringException {
        transformer = new MoveMethodTransformer(jmethod, refName, deleteParam, insertParam);
        transformer.changeSrc();  
        unparse(jfile);

        transformer.changeDst(dclass);
        unparse(dclass.getJavaFile());
    }

    public void additionalTransformation() throws RefactoringException {
        ArrayList files = QueryProject.getAllJavaFilesCallingTheMethod(jmethod);
        Iterator it = files.iterator();
        while (it.hasNext()) {
            JavaFile jf = (JavaFile)it.next();

            if (jmethod.isStatic()) {
                transformer.changeRef(jf);
            } else {
                CodeHighlighter.run(jf, jmethod);
            }
            unparse(jf);
        }
    }

    public String getLog() {
        return jmethod.getName() + " from " + jclass.getName() + " to " + dclass.getName();
    }

    private void checkAbstract() throws RefactoringException {
        if (jmethod.isAbstract()) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " is abstract or declared in an interface");
        }
    }

    private void checkNative() throws RefactoringException {
        if (jmethod.isNative()) {
            throw new RefactoringException("Not rename:" +
              " the method " + jmethod.getName() + " is native");
        }
    }

    private void checkSynchronized() throws RefactoringException {
        if (jmethod.isSynchronized()) {
            throw new RefactoringException("Not rename:" +
              " the method " + jmethod.getName() + " is synchronized");
        }
    }

    private void checkMethodCall() throws RefactoringException {
        Iterator it = jmethod.getMethodCallNodes().iterator();
        while (it.hasNext()) {
            JavaCall ja = new JavaCall((Node)it.next());

            JavaMethod jm = ja.getCalledJavaMethod();
            if (jm != null && !QueryMethod.isAccessible(dclass, jm)) {
                throw new RefactoringException("Not move:" +
                  " the method " + jmethod.getName() + " will not access to the method " + jm.getName());
            }
        }
    }

    private void checkFieldAccess() throws RefactoringException {
        Iterator it = jmethod.getVariableAccessNodes().iterator();
        while (it.hasNext()) {
            JavaVariable jv = new JavaVariable((Node)it.next());

            JavaField jf = jv.getUsedJavaField();
            if (jf != null && !QueryField.isAccessible(dclass, jf)) {
                throw new RefactoringException("Not move:" +
                  " the method " + jmethod.getName() + " will not access to the field " + jf.getName());
                
            }
        }
    }

    private void checkCallWithSuper() throws RefactoringException {
        Iterator it = jmethod.getExpressionNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            JavaExpr jr = new JavaExpr(elem);

            if (jr.isMethodCall()) {
                Node prefix = jr.getDotPrefix();

                if (prefix != null && prefix.getNodeName().equals(JXConstants.KeywordElem)) {
                    String keyword = prefix.getFirstChild().getNodeValue();
                    if (keyword.equals("super")) {
                        
                        JavaCall ja = new JavaCall(elem);
                        if (ja.getRefClass().equals(jmethod.getJavaClass().getSuperClassName())) {

                        throw new RefactoringException("Not rename:" +
                          " the method " + jmethod.getName() + " calls a method with the \"super\" prefix");
                        }
                    }
                }
            }
        }
    }

    private void checkThisSuffix() throws RefactoringException {
        Iterator it = QueryMethod.getJavaExprThis(jmethod).iterator();
        while (it.hasNext()) {
            JavaExpr jr = (JavaExpr)it.next();

            if (jr.isDotChild()) {
                Element opelem = (Element)jr.getNextSibling();
                if (opelem == null || !opelem.getNodeName().equals(JXConstants.OperatorElem) ||
                    !opelem.getFirstChild().getNodeValue().equals(".")) {
                    throw new RefactoringException("Not rename:" +
                      " the method " + jmethod.getName() + " includes the \"this\" suffix");
                }
            }
        }
    }

    private void checkParameterAndField() throws RefactoringException {
        ArrayList params = collectParametersOfTheClass(jmethod, null);
        ArrayList fields = collectFieldsOfTheClass(jmethod.getJavaClass(), null);
        if (params.size() == 0 && fields.size() == 0) {
            throw new RefactoringException("Not move:" +
              " a candidate of the target class is not found. An instance method can be moved to class used as a parameters or field");
        }
    }

    private void checkOverriddenOrOverridingMethod() throws RefactoringException {
        if (QueryMethod.getJavaMethodsWithTheSameSignature(jmethod, jclass.getAncestors()).size() != 1) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " overrides a method in ancestor of the source class " + jclass.getName()); 
        }

        if (QueryMethod.getJavaMethodsWithTheSameSignature(jmethod, jclass.getDescendants()).size() != 1) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " is overridden by a method in descendants of the source class " + jclass.getName());
        }
    }

    private void checkTargetClass() throws RefactoringException {
        if (dclass.isInterface()) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " can not be moved to the interface " + dclass.getName());
        }

        if (jmethod.isStatic() &&jmethod.isPublic() && !dclass.isPublic() &&
            !dclass.getJavaFile().getBaseName().equals(dclass.getName() + ".java")) {
            throw new RefactoringException("Not move:" +
              " the public method " + jmethod.getName() + " can be moved to a class defined in the file " + dclass.getName() + ".java");
        }
    }

    private void checkMovedMethod() throws RefactoringException {
        if (jclass.getAncestors().contains(dclass)) {
            throw new RefactoringException("Not move:" +
              " the class " + dclass.getName() + " is an ancestor of the source class " + jclass.getName() + ".\n" +
              "Do \"Pull Up Method\"!");
        }

        if (jclass.getDescendants().contains(dclass)) {
            throw new RefactoringException("Not move:" +
              " the class " + dclass.getName() + " is a descendant of the source class " + jclass.getName() + ".\n" +
              "Do \"Push Down Method\"!");
        }

        if (QueryClass.hasTheMethod(dclass, jmethod)) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " already exists in the target class " + dclass.getName());
        }

        if (!QueryClass.isAccessible(dclass, jmethod.getUsedTypes())) {
            checkInnerClass(jmethod);

            throw new RefactoringException("Not move:" +
              " the target class " + dclass.getName() + " will not access to the classes used in the moved method " + jmethod.getName());
        }

        if (jmethod.isStatic()) {
            Iterator it = dclass.getAllMethods().iterator();
            while (it.hasNext()) {
                JavaMethod jm = (JavaMethod)it.next();
                
                if (!jm.isStatic() && jmethod.getName().equals(jm.getName())) {
                    throw new RefactoringException("Not move:" +
                      " the instance method " + jm.getName() + " already exists in the target class " + dclass.getName());
                }
            }
        }
    }

    private void checkInnerClass(JavaMethod jm) throws RefactoringException {
        Iterator it = jm.getUsedTypes().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            if (jc.isInnerClass()) {
               throw new RefactoringException("Not move:" +
                 " the method " + jm.getName() + " uses an inner class " + jc.getName());
            }
        }
    }            

    private void determineReferenceName() throws RefactoringException {
        if (jmethod.isStatic()) {
            refName = dclass.getName();

            Iterator it = QueryProject.getJavaClassesCallingTheMethod(jmethod).iterator();
            while (it.hasNext()) {
                JavaClass jc = (JavaClass)it.next();

                if (!QueryClass.isAccessible(jc, dclass)) {
                    throw new RefactoringException("Not move:" +
                      " the class " + jc.getName() + " does not access to the target class " + dclass.getName());
                }
            }

        } else {
            if (!QueryClass.isAccessible(jclass, dclass)) {
                throw new RefactoringException("Not move:" +
                  " the class " + jclass.getName() + " does not access to the target class " + dclass.getName());
            }

            if (!QueryClass.isAccessible(jclass, dclass, jmethod)) {
                throw new RefactoringException("Not move:" +
                  " the class " + jclass.getName() + " will not access to the moved method " + jmethod.getName());
            }

            refName = getReferenceName();
             
            if (QueryMethod.getJavaExprThis(jmethod).size() != 0) {
                insertParam = true;
            }

            JavaMethodList ml = QueryMethod.getJavaMethodsCallingMethodsInAncestorsOfTheClass(jmethod, jclass);
            ml.remove(jmethod);
            if (ml.size() != 0) {
                if (!jmethod.isStatic()) {
                    insertParam = true;
                }
            }

            JavaFieldList fl = QueryMethod.getJavaFieldsUsingFieldsInAncestorsOfTheClass(jmethod, jclass);
            if (fl.size() != 0) {
                if (deleteParam) {
                    insertParam = true;
                } else {
                    Iterator it = fl.iterator();
                    while (it.hasNext()) {
                        JavaField jf = (JavaField)it.next();
                        if (!jf.getName().equals(refName)) {
                            insertParam = true;
                        }
                    }
                }
            }
        }         
    }

    private String getReferenceName() throws RefactoringException {
        Iterator it = QueryMethod.getWriteJavaFieldsUsingFieldsInAncestorsOfTheClass(jmethod, jclass).iterator();
        while (it.hasNext()) {
            JavaField jf = (JavaField)it.next();

            if (jf.getTypeFqn().equals(dclass.getFqn())) {
                throw new RefactoringException("Not move:" +
                  " the method " + jmethod.getName() + " changes a field " + jclass.getName() + " indicating the target class " + dclass.getName());
            }
        }

        JavaFieldList readFields = new JavaFieldList();
        it = QueryMethod.getReadJavaFieldsUsingFieldsInAncestorsOfTheClass(jmethod, jclass).iterator();
        while (it.hasNext()) {
            JavaField jf = (JavaField)it.next();

            /*
            if (!QueryField.isAccessible(dclass, jf)) {
                throw new RefactoringException("Not move:" +
                  " the method " + jmethod.getName() + " uses a field not accessible from " + dclass.getName());
            }
            */
            if (jf.getTypeFqn().equals(dclass.getFqn())) {
                readFields.add(jf);
            }
        }

        /*
        if (readFields.size() > 1) {
            throw new RefactoringException("Not move:" +
              " the method " + jmethod.getName() + " uses multiple fields in the source class " + jclass.getName());
        }
        */

        if (readFields.size() == 1) {
            JavaField jf = (JavaField)readFields.get(0);
            if (!jf.getTypeFqn().equals(dclass.getFqn())) {
                throw new RefactoringException("Not move:" +
                  " the method " + jmethod.getName() + " uses a field the type of which is not " + dclass.getName());
            }
            return jf.getName();
         
        } else {
            ArrayList fields = collectFieldsOfTheClass(jclass, dclass);
            ArrayList params = collectParametersOfTheClass(jmethod, dclass);
            
            if (fields.size() == 0 && params.size() == 0) {
                if (QueryProject.getAllJavaFilesCallingTheMethod(jmethod).size() == 0) {
                    return null;

                } else {
                    throw new RefactoringException("Not move:" +
                      " no reference to the target class " + dclass.getName() + " can be specified");
                }
            }

            if (fields.size() == 1 && params.size() == 0) {
                JavaField f = jclass.getJavaField((String)fields.get(0));
                return f.getName();
            }

            if (params.size() == 1) {
                JavaParameter p = jmethod.getParameterNode((String)params.get(0));
                deleteParam = true;
                return p.getName();

            }

            if (refName != null) {  // for testing
                JavaParameter p = jmethod.getParameterNode(refName);
                if (p != null) {
                    deleteParam = true;
                    return refName;
                }

                JavaField f = jclass.getJavaField(refName);
                if (f != null) {
                    return refName;
                }

            } else {

                String names[] = new String[fields.size() + params.size()];
                for (int i = 0; i < fields.size(); i++) {
                    names[i] = (String)fields.get(i) + " # field";
                }
                for (int i = 0; i < params.size(); i++) {
                    names[i + fields.size()] = (String)params.get(i) + " # parameter";
                }
                
                ReferenceDialog dialog = new ReferenceDialog(frame, "Reference name", dclass.getName(), names);
                String name = dialog.getRefName();
                if (name.endsWith(" # parameter")) {
                    deleteParam = true;
                }
                return name.substring(0, name.lastIndexOf("#") - 1);
            }
        }
        return null;
    }

    private ArrayList collectFieldsOfTheClass(JavaClass jc, JavaClass target) {
        ArrayList names = new ArrayList();

        Iterator it = jc.getAllFields().iterator();
        while (it.hasNext()) {
            JavaField jf = (JavaField)it.next();

            if (QueryProject.getJavaClassInProject(jf.getTypeFqn()) != null) {
                if (target == null || jf.getTypeFqn().equals(target.getFqn())) {
                    names.add(jf.getName());
                }
            }
        }

        return names;
    }

    private ArrayList collectParametersOfTheClass(JavaMethod jm, JavaClass target) {
        ArrayList names = new ArrayList();

        Iterator it = jm.getParameterNodes().iterator();
        while (it.hasNext()) {
            JavaParameter jp = new JavaParameter((Node)it.next());

            if (QueryProject.getJavaClassInProject(jp.getTypeFqn()) != null) {
                if (target == null || jp.getTypeFqn().equals(target.getFqn())) {
                    names.add(jp.getName());
                }
            }
        }

        return names;
    }
}
