/*
 *     RenameMethod.java  Aug 15, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.methods;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.jrbx.refactor.util.FakeJavaMethod;
import org.jtool.platform.xml.wrapper.*;
import java.util.*;
import javax.swing.JFrame;

public class RenameMethod extends MethodRefactoring {
    private String newName = null;
    private boolean updateReferences = true;
    private RenameMethodTransformer transformer;
    private JavaMethodList relativeMethods = new JavaMethodList();
    private String[] specialSignatures = { "toString( )", "equals( Object )", "equals( java.lang.Object )",
                                           "getClass( )", "hashCode( )", "notify( )", "notifyAll( )",
                                           "wait( )", "wait( long )", "wait( long int )" };

    public RenameMethod() {
        super();
    }

    public void setUpWithOptions(JFrame f, String newName, boolean ref) {
        frame = f;
        this.newName = newName;
        updateReferences = ref;
    }

    public void preconditions() throws RefactoringException {
        checkNative(jmethod);
        checkSpecialName(jmethod);
        overridesAbstractMethod();
        implementsMethod();

        if (newName == null) {
            RenameDialog dialog = new RenameDialog(frame, "Rename Method", jmethod.getName());
            newName = dialog.getNewName();

            if (newName == null) {
                throw new RefactoringException("");
            }
            updateReferences = dialog.getUpdateCheckBox();
        }

        FakeJavaMethod fjm = new FakeJavaMethod(jmethod);
        fjm.setName(newName);
        checkRenamedMethod(fjm);
        checkMethodInOuterClass(fjm);
        checkMethodInInnerClass(jmethod, fjm);

        if (jclass.isInterface()) {
            checkSpecialName(fjm);

            relativeMethods = getRelativeMethodsI(jmethod);
        } else {
            relativeMethods = getRelativeMethods(jmethod);
        }
        relativeMethods.add(jmethod);
        relativeMethods.addAll(getRelativeMethodsInAnonymousClass(relativeMethods));
        relativeMethods.remove(jmethod);

        Iterator it = relativeMethods.iterator();
        while (it.hasNext()) {
            JavaMethod jm = (JavaMethod)it.next();

            checkNative(jm);

            fjm = new FakeJavaMethod(jm);
            fjm.setName(newName);
            checkRenamedMethod(fjm);
            checkMethodInOuterClass(fjm);
            checkMethodInInnerClass(jm, fjm);
        }
    }

    public void transform() throws RefactoringException {
        transformer = new RenameMethodTransformer(jmethod, newName);
        transformer.changeSrc();
        unparse(jfile);
    }

    public void additionalTransformation() throws RefactoringException {
        if (!updateReferences) {
            return;
        }

        changeRef(jmethod);

        Iterator it = relativeMethods.iterator();
        while (it.hasNext()) {
            JavaMethod jm = (JavaMethod)it.next();

            transformer.changeRel(jm);
            unparse(jm.getJavaFile());

            changeRef(jm);
        }
    }

    private void changeRef(JavaMethod jm) {
        ArrayList files = QueryProject.getAllJavaFilesCallingTheMethod(jm);
        Iterator it = files.iterator();
        while (it.hasNext()) {
            JavaFile jf = (JavaFile)it.next();

            transformer.changeRef(jf, jm);
            unparse(jf);
        }
    }

    public String getLog() {
        return jmethod.getName() + " to " + newName;
    }

    private void overridesAbstractMethod() throws RefactoringException {
        Iterator it = QueryMethod.getOverriddenJavaMethods(jmethod).iterator();
        while (it.hasNext()) {
            JavaMethod jm = (JavaMethod)it.next();

            if (!jm.isAbstract()) {
                throw new RefactoringException("Not rename:" +
                  " the method " + jmethod.getName() + " overeides a method in " + jm.getJavaClass().getName());
            }
        }
    }

    private void implementsMethod() throws RefactoringException {
        if (!jmethod.getJavaClass().isInterface() && QueryMethod.getIndirectlyImplementedJavaMethods(jmethod).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the method " + jmethod.getName() + " is the implementation of a method");
        }
    }

    private void checkNative(JavaMethod jm) throws RefactoringException {
        if (jm.isNative()) {
            throw new RefactoringException("Not rename:" +
              " the method " + jm.getName() + " or method to be renamed together is native");
        }
    }

    private void checkSpecialName(JavaMethod jm) throws RefactoringException {
        String signature = jm.getSignature();
        for (int i = 0; i < specialSignatures.length; i++) {
            if (signature.equals(specialSignatures[i])) {
                throw new RefactoringException("Not rename:" +
                  " the method " + jm.getName() + " is special name " + specialSignatures[i]);
            }
        }
    }

    private JavaMethodList getRelativeMethods(JavaMethod jm) {
        JavaMethodList methods = new JavaMethodList();
        collectRelativeMethods(methods, jm);
        return methods;
    }

    private void collectRelativeMethods(JavaMethodList methods, JavaMethod jm) {
        JavaMethodList ms = QueryMethod.getOverriddenJavaMethods(jm);
        ms.addAll(QueryMethod.getOverridingJavaMethods(jm));
        
        Iterator it = ms.iterator();
        while (it.hasNext()) {
            JavaMethod m = (JavaMethod)it.next();
            if (!methods.contains(m)) {
                methods.add(m);
                collectRelativeMethods(methods, m);
            }
        }
    }

    private JavaMethodList getRelativeMethodsI(JavaMethod jm) {
        JavaMethodList methods = new JavaMethodList();
        collectRelativeMethodsI(methods, jm);
        return methods;
    }

    private void collectRelativeMethodsI(JavaMethodList methods, JavaMethod jm) {
        JavaMethodList ms = QueryMethod.getOverriddenJavaMethods(jm);
        ms.addAll(QueryMethod.getOverridingJavaMethods(jm));
        ms.addAll(QueryMethod.getIndirectlyImplementingJavaMethods(jm));
        ms.addAll(QueryMethod.getIndirectlyImplementedJavaMethods(jm));
        
        Iterator it = ms.iterator();
        while (it.hasNext()) {
            JavaMethod m = (JavaMethod)it.next();
            if (!methods.contains(m)) {
                methods.add(m);
                collectRelativeMethodsI(methods, m);
            }
        }
    }

    private JavaMethodList getRelativeMethodsInAnonymousClass(JavaMethodList methods) {
        JavaMethodList ms = new JavaMethodList();

        Iterator it = methods.iterator();
        while (it.hasNext()) {
            JavaMethod jm = (JavaMethod)it.next();

            if (!jm.isPrivate()) {
                Iterator it2 = QueryProject.getAnonymousJavaClasses(jm.getJavaClass()).iterator();
                while (it2.hasNext()) {
                    JavaClass jc = (JavaClass)it2.next();
                    
                    JavaMethod m = jc.getJavaMethod(jm.getSignature());
                    if (m != null) {
                        ms.add(m);
                    }
                }
            }
        }
        return ms;
    }

    public void checkRenamedMethod(JavaMethod jm) throws RefactoringException {
        JavaClass fjc = jm.getJavaClass();
        if (QueryClass.hasTheMethod(fjc, jm)) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " already exists in the class " + fjc.getName());
        }            

        if (QueryMethod.getOverriddenJavaMethods(jm).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " overrides a method in the ancestors of the method " + fjc.getName());
        }

        if (QueryMethod.getOverridingJavaMethods(jm).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " is overridden a method in the descendants of the method " + fjc.getName());
        }

        if (QueryMethod.getHiddenJavaMethods(jm).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " hides a method in the ancestors of the method " + fjc.getName());
        }

        if (QueryMethod.getHidingJavaMethods(jm).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the method " + newName + " is hidden a method in the descendants of the method " + fjc.getName());
        }

        if (QueryMethod.getApplicableJavaMethodsInAncestors(jm).size() != 0 ||
            QueryMethod.getApplicableJavaMethodsInDescendants(jm).size() != 0) {
            throw new RefactoringException("Not rename:" +
              " the applicable method " + newName + " already exists in the class " + fjc.getName());
        }
    }

    private void checkMethodInOuterClass(JavaMethod jm) throws RefactoringException {
        Iterator it = jm.getJavaClass().getDescendants().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            if (!jc.equals(jm.getClass())) {
                Iterator it2 = jc.getAllOuterClasses().iterator();
                while (it2.hasNext()) {
                    JavaClass jc2 = (JavaClass)it2.next();

                    if (!jc2.isInterface()) {
                        JavaMethod jm2 = jc2.getJavaMethod(jm.getSignature());
                    
                        if (jm2 != null && QueryClass.getJavaCallsUsingTheMethod(jc, jm2).size() != 0) {
                            throw new RefactoringException("Not rename:" +
                              " the method " + jm2.getName() + " of the class " + jc2.getName() + " will be shadowed");
                        }
                    }
                }
            }
        }
    }

    private void checkMethodInInnerClass(JavaMethod jm, JavaMethod newjm) throws RefactoringException {
        Iterator it = jm.getJavaClass().getAllInnerClassesAndInterfaces().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            if (!jc.isInterface()) {
                if (QueryClass.getJavaCallsUsingTheMethod(jc, jm).size() != 0) {

                    Iterator it2 = jc.getAncestors().iterator();
                    while (it2.hasNext()) {
                        JavaClass jc2 = (JavaClass)it2.next();

                        JavaMethod jm2 = jc2.getJavaMethod(newjm.getSignature());
                        if (jm2 != null && (!jm2.isPrivate() || jc.equals(jc2))) {
                            throw new RefactoringException("Not rename:" +
                              " the method " + jm2.getName() + " of the class " + jc2.getName() + " will shadow the renamed method");
                        }
                    }
                }
            }
        }
    }
}
